/*
 * ex: sw=4
 */

// #include <linux/config.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <net/sock.h>
#include <linux/rtnetlink.h>
#include <linux/wireless.h>
#include <net/iw_handler.h>

#include <hcf/debug.h>
#include <hcf.h>
#include <hcfdef.h>

#include <wireless/wl_if.h>
#include <wireless/wl_internal.h>
#include <wireless/wl_util.h>
#include <wireless/wl_main.h>
#include <wireless/wl_wext.h>
#include <wireless/wl_priv.h>

// #define to_net_dev(class) container_of(class, struct net_device, class_dev)

static inline int dev_isalive(const struct net_device *dev)
{
            return dev->reg_state == NETREG_REGISTERED;
}

/*
 * empirically even if tallies are defined as 32 bits entities, only
 * high 16 bits are relevant; low half is always zero. It means tallies
 * are pretty much useless for traffic counting but at least gice overview
 * about where error come from
 */
static ssize_t show_tallies(struct class_device *cd, char *buf)
{
    struct net_device *dev = to_net_dev(cd);
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    CFG_HERMES_TALLIES_STRCT tallies;
    ssize_t ret = -EINVAL;

    read_lock(&dev_base_lock);
    if (dev_isalive(dev)) {
	if (wl_lock(lp, &flags))
	    ret = -EBUSY;
	else {
	    if ((ret = wl_get_tallies(lp, &tallies)) == 0) {
		wl_unlock(lp, &flags);
		ret = snprintf(buf, PAGE_SIZE,
		    "TxUnicastFrames:           %u\n"
		    "TxMulticastFrames:         %u\n"
		    "TxFragments:               %u\n"
		    "TxUnicastOctets:           %u\n"
		    "TxMulticastOctets:         %u\n"
		    "TxDeferredTransmissions:   %u\n"
		    "TxSingleRetryFrames:       %u\n"
		    "TxMultipleRetryFrames:     %u\n"
		    "TxRetryLimitExceeded:      %u\n"
		    "TxDiscards:                %u\n"
		    "RxUnicastFrames:           %u\n"
		    "RxMulticastFrames:         %u\n"
		    "RxFragments:               %u\n"
		    "RxUnicastOctets:           %u\n"
		    "RxMulticastOctets:         %u\n"
		    "RxFCSErrors:               %u\n"
		    "RxDiscardsNoBuffer:        %u\n"
		    "TxDiscardsWrongSA:         %u\n"
		    "RxWEPUndecryptable:        %u\n"
		    "RxMsgInMsgFragments:       %u\n"
		    "RxMsgInBadMsgFragments:    %u\n"
		    "RxDiscardsWEPICVError:     %u\n"
		    "RxDiscardsWEPExcluded:     %u\n"
		    ,
		    (unsigned int)tallies.TxUnicastFrames >> 16,
		    (unsigned int)tallies.TxMulticastFrames >> 16,
		    (unsigned int)tallies.TxFragments >> 16,
		    (unsigned int)tallies.TxUnicastOctets >> 16,
		    (unsigned int)tallies.TxMulticastOctets >> 16,
		    (unsigned int)tallies.TxDeferredTransmissions >> 16,
		    (unsigned int)tallies.TxSingleRetryFrames >> 16,
		    (unsigned int)tallies.TxMultipleRetryFrames >> 16,
		    (unsigned int)tallies.TxRetryLimitExceeded >> 16,
		    (unsigned int)tallies.TxDiscards >> 16,
		    (unsigned int)tallies.RxUnicastFrames >> 16,
		    (unsigned int)tallies.RxMulticastFrames >> 16,
		    (unsigned int)tallies.RxFragments >> 16,
		    (unsigned int)tallies.RxUnicastOctets >> 16,
		    (unsigned int)tallies.RxMulticastOctets >> 16,
		    (unsigned int)tallies.RxFCSErrors >> 16,
		    (unsigned int)tallies.RxDiscardsNoBuffer >> 16,
		    (unsigned int)tallies.TxDiscardsWrongSA >> 16,
		    (unsigned int)tallies.RxWEPUndecryptable >> 16,
		    (unsigned int)tallies.RxMsgInMsgFragments >> 16,
		    (unsigned int)tallies.RxMsgInBadMsgFragments >> 16,
		    (unsigned int)tallies.RxDiscardsWEPICVError >> 16,
		    (unsigned int)tallies.RxDiscardsWEPExcluded >> 16);
	    }
	}
    }

    read_unlock(&dev_base_lock);
    return ret;
}

static CLASS_DEVICE_ATTR(tallies, S_IRUGO, show_tallies, NULL);

static struct attribute *wlags_attrs[] = {
    &class_device_attr_tallies.attr,
    NULL
};

static struct attribute_group wlags_group = {
    .name = "wlags",
    .attrs = wlags_attrs,
};

void register_wlags_sysfs(struct net_device *dev)
{
//    struct class_device *class_dev = &(dev->class_dev);
    struct wl_private *lp = wl_priv(dev);

//    lp->sysfsCreated = sysfs_create_group(&class_dev->kobj, &wlags_group);
}

void unregister_wlags_sysfs(struct net_device *dev)
{
//    struct class_device *class_dev = &(dev->class_dev);
    struct wl_private *lp = wl_priv(dev);

//    if (lp->sysfsCreated)
//	sysfs_remove_group(&class_dev->kobj, &wlags_group);
}
