/*******************************************************************************
 * ex: sw=4
 * Agere Systems Inc.
 * Wireless device driver for Linux (wlags49).
 *
 * Copyright (c) 1998-2003 Agere Systems Inc. 
 * All rights reserved.
 *   http://www.agere.com
 *
 * Initially developed by TriplePoint, Inc.
 *   http://www.triplepoint.com
 *
 *------------------------------------------------------------------------------
 *
 * SOFTWARE LICENSE
 *
 * This software is provided subject to the following terms and conditions,
 * which you should read carefully before using the software.  Using this
 * software indicates your acceptance of these terms and conditions.  If you do
 * not agree with these terms and conditions, do not use the software.
 *
 * Copyright  2003 Agere Systems Inc.
 * All rights reserved.
 *
 * Redistribution and use in source or binary forms, with or without
 * modifications, are permitted provided that the following conditions are met:
 *
 * . Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following Disclaimer as comments in the code as
 *    well as in the documentation and/or other materials provided with the
 *    distribution.
 * 
 * . Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following Disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * . Neither the name of Agere Systems Inc. nor the names of the contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * Disclaimer
 *
 * THIS SOFTWARE IS PROVIDED AS IS AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  ANY
 * USE, MODIFICATION OR DISTRIBUTION OF THIS SOFTWARE IS SOLELY AT THE USERS OWN
 * RISK. IN NO EVENT SHALL AGERE SYSTEMS INC. OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED TO, CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************************/




/*******************************************************************************
 * VERSION CONTROL INFORMATION
 *******************************************************************************
 *
 * $Author: vjs $
 * $Date: 2004/03/31 14:47:45 $
 * $Revision: 1.45 $
 * $Source: /cvsroot/wifi/wl_lkm/wireless/wl_wext.c,v $
 *
 ******************************************************************************/




/*******************************************************************************
 *  include files
 ******************************************************************************/
#include <wireless/wl_version.h>

#include <linux/if_arp.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <asm/uaccess.h>

#include <hcf/debug.h>
#include <hcf.h>
#include <hcfdef.h>

#include <wireless/wl_if.h>
#include <wireless/wl_internal.h>
#include <wireless/wl_util.h>
#include <wireless/wl_main.h>
#include <wireless/wl_wext.h>
#include <wireless/wl_priv.h>


/* If WIRELESS_EXT is not defined (as a result of HAS_WIRELESS_EXTENSIONS
   #including linux/wireless.h), then these functions do not need to be included
   in the build. */
#ifdef WIRELESS_EXT




/*******************************************************************************
 * global definitions
 ******************************************************************************/
// Frequency of channel 1..14 in MHz
static const long frequency_list[] = 
{ 
    2412, 2417, 2422, 2427, 2432, 2437, 2442,
    2447, 2452, 2457, 2462, 2467, 2472, 2484
};

static struct iw_statistics *wl_wireless_stats(struct net_device *);

#if DBG
extern dbg_info_t *DbgInfo;
#endif  // DBG

static int wl_siwcommit(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *rqu, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags))
       ret = -EBUSY;
    else {
	wl_apply(lp);
	wl_unlock(lp, &flags);
    }

    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwname(struct net_device *dev,
		       struct iw_request_info *info,
		       char *name, char *extra)
{
    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    strcpy(name, "IEEE 802.11b");

    DBG_LEAVE(DbgInfo);
    return 0;
}

static int wl_siwfreq(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_freq *freq, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int channel = -1;
    int ret     = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    /* If frequency specified, look up channel */
    if(freq->e == 1) {
        int f = freq->m / 100000;
        int i;

        for(i = 0; i < 14; i++)
            if(f == frequency_list[i]) {
                channel = i+1;
                break;
	}
    }

    /* Channel specified */
    if(freq->e == 0)
        channel = freq->m;

    /* Validate the new value against the valid adapter channels. */
    if(channel >= 0 && channel <= 14) {
#if (HCF_TYPE) & HCF_TYPE_HII
        /* Define a method for validating the channel selection for Hermes-II,
           as CFG_CHANNEL_LIST is only supported for Hermes-I */
#else
        __u16   *pChanList;
        int     status  = -1;

	if (wl_lock(lp, &flags)) {
	    ret = -EBUSY;
	    goto out;
	}

        lp->ltvRecord.len = 1 + (sizeof( *pChanList)/sizeof(hcf_16));
        lp->ltvRecord.typ = CFG_CHANNEL_LIST;
        status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));

        if(status == HCF_SUCCESS) {
            pChanList = (__u16 *)&(lp->ltvRecord.u.u32);
 
            /* Check if the channel is acceptable. */
            if(!((__le16_to_cpu(*pChanList)) & (0x01L << (channel - 1)))) {
                ret = -EINVAL;
		goto out_unlock;
            }
        } else {
            ret = -EFAULT;
	    goto out_unlock;
        }

#endif /* (HCF_TYPE) & HCF_TYPE_HII */    
    } else {
        ret = -EINVAL;
	goto out;
    }
 
    lp->Channel = channel;

    /* Commit the adapter parameters */
    wl_apply(lp);

    /* Send an event that channel/freq has been set */ 
    wl_wext_event_freq(lp->dev);

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;    
}

static int wl_giwfreq(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_freq *freq, char *extra)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int status = -1;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    lp->ltvRecord.len = 2;
    lp->ltvRecord.typ = CFG_CURRENT_CHANNEL;
    
    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
    if( status == HCF_SUCCESS ) {
        int channel = __le16_to_cpu(lp->ltvRecord.u.u16[0]);

#ifdef USE_FREQUENCY
        if( channel >= 1 && channel <= RADIO_CHANNELS ) {
             freq->m = frequency_list[channel-1] * 100000;
             freq->e = 1;
        }
#else
        freq->m = channel;
        freq->e = 0;
#endif /* USE_FREQUENCY */
    }

    wl_unlock(lp, &flags);
    ret = (status == HCF_SUCCESS ? 0 : -EFAULT);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_giwrange(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    struct iw_range *range = (struct iw_range *) extra;
    int ret = 0;
    int status = -1;
    int count;
    __u16 *pTxRate;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    /* Set range information */
    data->length = sizeof(struct iw_range);
    memset(range, 0, sizeof(struct iw_range));

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Get the current transmit rate from the adapter */
    lp->ltvRecord.len = 1 + (sizeof(*pTxRate) / sizeof(hcf_16));
    lp->ltvRecord.typ = CFG_CURRENT_TX_RATE;
    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
    
    if(status == HCF_SUCCESS) {
	pTxRate = (__u16 *)&(lp->ltvRecord.u.u32);
	range->throughput = __le16_to_cpu(*pTxRate) * MEGABIT;
    } else {
	ret = -EFAULT;
	goto out_unlock;
    }

    // NWID - NOT SUPPORTED

    /* Channel/Frequency Info */
    range->num_channels = RADIO_CHANNELS;

    /* Signal Level Thresholds */
    range->sensitivity = RADIO_SENSITIVITY_LEVELS;

    /* Link quality */
#ifdef USE_DBM
    range->max_qual.qual     = (u_char)HCF_MAX_COMM_QUALITY;
    
    /* If the value returned in /proc/net/wireless is greater than the maximum range,
       iwconfig assumes that the value is in dBm. Because an unsigned char is used,
       it requires a bit of contorsion... */

    range->max_qual.level    = (u_char)(dbm(HCF_MIN_SIGNAL_LEVEL) - 1);
    range->max_qual.noise    = (u_char)(dbm(HCF_MIN_NOISE_LEVEL) - 1);
#else
    range->max_qual.qual     = 100;
    range->max_qual.level    = 100;
    range->max_qual.noise    = 100;
#endif /* USE_DBM */

    /* Set available rates */
    range->num_bitrates = 0;

    lp->ltvRecord.len = 6;
    lp->ltvRecord.typ = CFG_SUPPORTED_DATA_RATES;
    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
	    
    if(status == HCF_SUCCESS) {
	for(count = 0; count < MAX_RATES; count++)
	    if(lp->ltvRecord.u.u8[count+2] != 0) {
		range->bitrate[count] = lp->ltvRecord.u.u8[count+2] * MEGABIT / 2;
		range->num_bitrates++;
	    }
    } else {
	ret = -EFAULT;
	goto out_unlock;
    }

    /* RTS Threshold info */
    range->min_rts   = MIN_RTS_BYTES;
    range->max_rts   = MAX_RTS_BYTES;

    // Frag Threshold info - NOT SUPPORTED

    // Power Management info - NOT SUPPORTED

    /* Encryption */

#if WIRELESS_EXT > 8
    /* Is WEP supported? */
    if(wl_has_wep(&(lp->hcfCtx))) {
	/* WEP: RC4 40 bits */
	range->encoding_size[0]      = MIN_KEY_SIZE;

	/* RC4 ~128 bits */
	range->encoding_size[1]      = MAX_KEY_SIZE;
	range->num_encoding_sizes    = 2;
	range->max_encoding_tokens   = MAX_KEYS;
    }
#endif /* WIRELESS_EXT > 8 */

    /* Tx Power Info */
    range->txpower_capa  = IW_TXPOW_MWATT;
    range->num_txpower   = 1;
    range->txpower[0]    = RADIO_TX_POWER_MWATT;

#if WIRELESS_EXT > 10
    /* Wireless Extension Info */
    range->we_version_compiled   = WIRELESS_EXT;
    range->we_version_source     = WIRELESS_SUPPORT; 

    // Retry Limits and Lifetime - NOT SUPPORTED
#endif


#if WIRELESS_EXT > 11
    wl_wireless_stats(lp->dev);
    range->avg_qual = lp->wstats.qual;
#endif

    /* Event capability (kernel + driver) */
    range->event_capa[0] = (IW_EVENT_CAPA_K_0 |
			    IW_EVENT_CAPA_MASK(SIOCGIWAP) |
			    IW_EVENT_CAPA_MASK(SIOCGIWSCAN));
    range->event_capa[1] = IW_EVENT_CAPA_K_1;
    range->event_capa[4] = (IW_EVENT_CAPA_MASK(IWEVREGISTERED) |
			    IW_EVENT_CAPA_MASK(IWEVCUSTOM) |
			    IW_EVENT_CAPA_MASK(IWEVEXPIRED));

    range->enc_capa = IW_ENC_CAPA_WPA | IW_ENC_CAPA_CIPHER_TKIP;

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwap(struct net_device *dev,
		    struct iw_request_info *info,
		    struct sockaddr *ap_addr, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    ap_addr->sa_family = ARPHRD_ETHER;

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Assume AP mode here, which means the BSSID is our own MAC address. In
       STA mode, this address will be overwritten with the actual BSSID using
       the code below. */
    memcpy(&ap_addr->sa_data, lp->dev->dev_addr, ETH_ALEN);

#if (HCF_TYPE) & HCF_TYPE_STA
    if( lp->DownloadFirmware != WVLAN_DRV_MODE_AP) {
	int status = -1;

        /* Get Current BSSID */
        lp->ltvRecord.typ = CFG_CURRENT_BSSID;
        lp->ltvRecord.len = 4;
        status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));

        if(status == HCF_SUCCESS)
            /* Copy info into sockaddr struct */
            memcpy(&ap_addr->sa_data, lp->ltvRecord.u.u8, ETH_ALEN);
        else
            ret = -EFAULT;
    }
#endif  /* (HCF_TYPE) & HCF_TYPE_STA */

    wl_unlock(lp, &flags);
 
out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwaplist (struct net_device *dev,
		  struct iw_request_info *info,
		  struct iw_point *data, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    int                 ret = 0;
    unsigned long	flags;
    int                 status = -1;
    int                 num_aps = -1;
    int                 sec_count = 0;
    hcf_32              count;
    struct sockaddr     *hwa = NULL;
    struct iw_quality   *qual = NULL;
    /*------------------------------------------------------------------------*/


    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }


    /* Set the completion state to FALSE */
    lp->scan_results.scan_complete = FALSE;
    lp->probe_results.scan_complete = FALSE;


    /* Channels to scan */
    lp->ltvRecord.len       = 2;
    lp->ltvRecord.typ       = CFG_SCAN_CHANNELS_2GHZ;
    lp->ltvRecord.u.u16[0]  = CNV_INT_TO_LITTLE( 0x7FFF );

    status = hcf_put_info( &( lp->hcfCtx ), (LTVP)&( lp->ltvRecord ));
    DBG_TRACE( DbgInfo, "CFG_SCAN_CHANNELS_2GHZ result: 0x%x\n", status );


    /* Set the SCAN_SSID to "ANY". Using this RID for scan prevents the need to
       disassociate from the network we are currently on */
    lp->ltvRecord.len       = 2;
    lp->ltvRecord.typ       = CFG_SCAN_SSID;
    lp->ltvRecord.u.u16[0]  = CNV_INT_TO_LITTLE( 0 );
    
    status = hcf_put_info( &( lp->hcfCtx ), (LTVP)&( lp->ltvRecord ));
    DBG_TRACE( DbgInfo, "CFG_SCAN_SSID to 'any' status: 0x%x\n", status );


#if (HCF_TYPE) & HCF_TYPE_HII
    /* Initiate the scan */
    status = hcf_action( &( lp->hcfCtx ), HCF_ACT_ACS_SCAN );
#else
    status = hcf_action( &( lp->hcfCtx ), HCF_ACT_SCAN );
#endif  /* (HCF_TYPE) & HCF_TYPE_HII */

    wl_unlock(lp, &flags);

    if(status == HCF_SUCCESS) {
        DBG_TRACE( DbgInfo, "SUCCESSFULLY INITIATED SCAN...\n" );

        do {
            DBG_TRACE( DbgInfo, "Waiting for scan results...\n" );

            /* Abort the scan if we've waited for more than MAX_SCAN_TIME_SEC */
            if(sec_count > MAX_SCAN_TIME_SEC) {
                DBG_ERROR( DbgInfo, "Scan did not complete in allotted time\n" );
                ret = -EIO;
                break;
            }

            /* Wait for 1 sec in 10ms intervals, scheduling the kernel to do
               other things in the meantime, This prevents system lockups by
               giving some time back to the kernel */
            for(count = 0; count < 100; count ++) {
                mdelay(10);
                schedule( );
            }

            /* We've waited for another second, so increase the second count */
            sec_count++;
        }
#if (HCF_TYPE) & HCF_TYPE_HII
        while( lp->probe_results.scan_complete == FALSE );
#else
        while( lp->scan_results.scan_complete == FALSE );
#endif  /* (HCF_TYPE) & HCF_TYPE_HII */


	rmb();

#if (HCF_TYPE) & HCF_TYPE_HII
        if( lp->probe_results.scan_complete == FALSE )
#else
        if( lp->scan_results.scan_complete == FALSE )
#endif  /* (HCF_TYPE) & HCF_TYPE_HII */
        {
            DBG_ERROR( DbgInfo, "Unknown Error, timeout waiting for scan results? \n" );
            ret = -EIO;
        } else {
#if (HCF_TYPE) & HCF_TYPE_HII
            num_aps             = lp->probe_results.num_aps;
	    if (num_aps > IW_MAX_AP)
		num_aps = IW_MAX_AP;
#ifdef HERMES2
	    data->length = num_aps;
#else
	    wrq->u.data.length  = num_aps; */
#endif //HERMES2

            hwa = (struct sockaddr *)extra;
            qual = (struct iw_quality *)(extra + 
                                        sizeof(struct sockaddr) * num_aps);

            /* This flag is used to tell the user if we provide quality
               information. Since we provide signal/noise levels but no
               quality info on a scan, this is set to 0. Setting to 1 and 
               providing a quality of 0 produces weird results. If we ever
               provide quality (or can calculate it), this can be changed */
            data->flags = 0;

            for(count = 0; count < num_aps; count++) {
                DBG_PRINT("BSSID: %s\n", DbgHwAddr(lp->probe_results.ProbeTable[count].BSSID));
                memcpy(hwa[count].sa_data, 
                        lp->probe_results.ProbeTable[count].BSSID, ETH_ALEN);
            }

            /* Once the data is copied to the wireless struct, invalidate the
               scan result to initiate a rescan on the next request */
            lp->probe_results.scan_complete = FALSE;

#else
            /* We got an indication that the scan has completed, so fill out the
               wireless struct */
            num_aps = lp->scan_results.num_aps;
	    if (num_aps > IW_MAX_AP)
		num_aps = IW_MAX_AP;
            data->length = num_aps;

            hwa = (struct sockaddr *)extra;
            qual = (struct iw_quality *)(extra + 
                                        sizeof(struct sockaddr) * num_aps);

            /* This flag is used to tell the user if we provide quality
               information. Since we provide signal/noise levels but no
               quality info on a scan, this is set to 0. Setting to 1 and 
               providing a quality of 0 produces weird results. If we ever
               provide quality (or can calculate it), this can be changed */
            data->flags = 0;

            for(count = 0; count < num_aps; count++) {
                memcpy(hwa[count].sa_data, lp->scan_results.APTable[count].bssid, ETH_ALEN);
            }

            /* Once the data is copied to the wireless struct, invalidate the
               scan result to initiate a rescan on the next request */
            lp->scan_results.scan_complete = FALSE;

#endif  /* (HCF_TYPE) & HCF_TYPE_HII */

            /* Send the wireless event that the scan has completed, just in case
               it's needed */
            wl_wext_event_scan_complete(lp->dev);
        }
    }


out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_siwsens(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *sens, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if(sens->value < 1 || sens->value > 3) {
        ret = -EINVAL;
	goto out;
    }
	
    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    lp->DistanceBetweenAPs = sens->value;
    wl_apply(lp);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwsens(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *sens, char *extra)
{
    struct wl_private *lp = wl_priv(dev);

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    /* not worth locking ... */
    sens->value = lp->DistanceBetweenAPs;
    sens->fixed = 0;	/* auto */

    DBG_LEAVE(DbgInfo);
    return 0;
}

static int wl_siwessid(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *ssid)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (data->flags != 0 && data->length > HCF_MAX_NAME_LEN + 1) {
	ret = -EINVAL;
	goto out;
    }

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    memset(lp->NetworkName, 0, sizeof(lp->NetworkName));
    if(data->flags == 0)
        /* Need this because in STAP build PARM_DEFAULT_SSID is "LinuxAP" */
        if( lp->DownloadFirmware == WVLAN_DRV_MODE_STA )
            strcpy(lp->NetworkName, "ANY");
        else
            strcpy(lp->NetworkName, PARM_DEFAULT_SSID);
    else
        memcpy(lp->NetworkName, ssid, data->length);
	    
    /* Commit the adapter parameters */
    wl_apply(lp);

    /* Send an event that ESSID has been set */
    wl_wext_event_essid(lp->dev);

    wl_unlock(lp, &flags);

    DBG_TRACE(DbgInfo, "set NetworkName: %s\n", ssid);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwessid(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *essid)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;
    int status = -1;
    wvName_t    *pName;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Get the desired network name */
    lp->ltvRecord.len = 1 + (sizeof(*pName)/sizeof(hcf_16));

#if (HCF_TYPE) & HCF_TYPE_STA
    lp->ltvRecord.typ = CFG_DESIRED_SSID;
#endif

#if (HCF_TYPE) & HCF_TYPE_AP
    if(lp->DownloadFirmware == WVLAN_DRV_MODE_AP)
        lp->ltvRecord.typ = CFG_CNF_OWN_SSID;
#endif // HCF_AP

    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
    if(status == HCF_SUCCESS)
	pName = (wvName_t *)&(lp->ltvRecord.u.u32);
    else {
	ret = -EFAULT;
	goto out_unlock;
    }

#if (HCF_TYPE) & HCF_TYPE_STA
#ifdef RETURN_CURRENT_NETWORKNAME
    /* if desired is null ("any"), return current or "any" */
    if(pName->name[0] == '\0') {
	/* Get the current network name */
	lp->ltvRecord.len = 1 + (sizeof(*pName)/sizeof(hcf_16));
	lp->ltvRecord.typ = CFG_CURRENT_SSID;
	status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
	if(status == HCF_SUCCESS)
	    pName = (wvName_t *)&(lp->ltvRecord.u.u32);
	else {
	    ret = -EFAULT;
	    goto out_unlock;
	}
    }
#endif // RETURN_CURRENT_NETWORKNAME
#endif // HCF_STA

    /* Endian translate the string length */
    pName->length = __le16_to_cpu(pName->length); 
    if (pName->length > IW_ESSID_MAX_SIZE) {
	ret = -EFAULT;
	goto out_unlock;
    }

    /* Copy the information into the user buffer */
    data->flags = 1; /* FIXME is it slways true? */
    data->length = pName->length;
    memcpy(essid, pName->name, pName->length);

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_siwencode(struct net_device *dev,
			struct iw_request_info *info,
			struct iw_point *erq, char *keybuf)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int     ret = 0;
    int index;
    hcf_8 tk;
    hcf_8 encryption_state;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Is encryption supported? */
	if(!wl_has_wep(&(lp->hcfCtx))) {
        DBG_WARNING( DbgInfo, "WEP not supported on this device\n" );
        ret = -EOPNOTSUPP;
        goto out_unlock;
    }
    
    DBG_NOTICE(DbgInfo, "length: %d, flags: %#x\n", erq->length, erq->flags);

    /* Save state of Encryption switch */
    encryption_state = lp->EnableEncryption;

    /* Basic checking: do we have a key to set? */
    index  = (erq->flags & IW_ENCODE_INDEX) - 1;
    tk     = lp->TransmitKeyID - 1;		// current key

	    
    /* Check the size of the key */
    switch(erq->length) {
    case 0:
        /* Do we want to just set the current transmit key? */
	if (erq->flags & IW_ENCODE_DISABLED)
	    ; /* fall through */
	else if(index >= 0 && index < MAX_KEYS) {
            DBG_NOTICE(DbgInfo, "index: %d; len: %d\n", index,
                       lp->DefaultKeys.key[index].len);

            if(lp->DefaultKeys.key[index].len > 0) {
                lp->TransmitKeyID       = index + 1;
                lp->EnableEncryption    = 1;
            } else {
                DBG_WARNING( DbgInfo, "Problem setting the current TxKey\n" );
                DBG_LEAVE( DbgInfo );
		ret = -EINVAL;
		goto out_unlock;
	    }
        } else {
	    ret = -EINVAL;
	    goto out_unlock;
	}

	break;

    case MIN_KEY_SIZE:
    case MAX_KEY_SIZE:

	/* Check the index */
	if(index < 0 || index >= MAX_KEYS)
	    index = tk;

	/* Cleanup */
	memset(lp->DefaultKeys.key[index].key, 0, MAX_KEY_SIZE);
	    
	/* Copy the key in the driver */
	memcpy(lp->DefaultKeys.key[index].key, keybuf, erq->length);
	/* Set the length */
	lp->DefaultKeys.key[index].len = erq->length;

	DBG_NOTICE( DbgInfo, "encoding.length: %d\n", wrq->u.encoding.length );
	DBG_NOTICE( DbgInfo, "set key: %s(%d) [%d]\n", lp->DefaultKeys.key[index].key,
		    lp->DefaultKeys.key[index].len, index );

	/* Enable WEP (if possible) */
	if(index == tk && lp->DefaultKeys.key[tk].len > 0)
	    lp->EnableEncryption = 1;

	break;

    default:
	DBG_WARNING( DbgInfo, "Invalid Key length\n" );
	ret = -EINVAL;
	goto out_unlock;
    }

    /* Read the flags */ 
    if(erq->flags & IW_ENCODE_DISABLED)
        lp->EnableEncryption = 0;	// disable encryption
    else
        lp->EnableEncryption = 1;

    if(erq->flags & IW_ENCODE_RESTRICTED) {
        DBG_WARNING( DbgInfo, "IW_ENCODE_RESTRICTED invalid\n" );
	ret = -EINVAL;		// Invalid
	goto out_unlock;
    }
    
    DBG_TRACE(DbgInfo, "encryption_state:       %d\n", encryption_state);
    DBG_TRACE(DbgInfo, "lp->EnableEncryption:   %d\n", lp->EnableEncryption);
    DBG_TRACE(DbgInfo, "erq->length:            %d\n", erq->length);
    DBG_TRACE(DbgInfo, "erq->flags:           0x%x\n", erq->flags);

    /* Write the changes to the card */
    DBG_NOTICE(DbgInfo, "encrypt: %d, ID: %d\n", lp->EnableEncryption,
		lp->TransmitKeyID );

    if(lp->EnableEncryption == encryption_state) {
	if(erq->length || lp->TransmitKeyID != tk)
	    /* Dynamic WEP key update */
	    wl_set_wep_keys(lp);
    } else
	/* To switch encryption on/off, soft reset is required */
	wl_apply(lp);
    
    /* Send an event that Encryption has been set */
    wl_wext_event_encode(dev);

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_giwencode(struct net_device *dev,
		 struct iw_request_info *info,
		 struct iw_point *erq, char *key)

{
    struct wl_private *lp = wl_priv(dev);
    int ret = 0;
    int index = (erq->flags & IW_ENCODE_INDEX ) - 1;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );
    DBG_NOTICE(DbgInfo, "GIWENCODE: encrypt: %d, ID: %d\n", lp->EnableEncryption, lp->TransmitKeyID);
    

    /* Is it supported? */
    if(!wl_has_wep( &( lp->hcfCtx))) {
        ret = -EOPNOTSUPP;
        DBG_LEAVE( DbgInfo );
        return ret;
    }
	
    /* Only super-user can see WEP key */
    if(!capable(CAP_NET_ADMIN)) {
        ret = -EPERM;
        DBG_LEAVE( DbgInfo );
        return ret;
    }
	
    /* Set the flags */
    erq->flags = 0;
    
    if(lp->EnableEncryption == 0)
	erq->flags |= IW_ENCODE_DISABLED;
	     
    /* Which key do we want */
    if(( index < 0 ) || ( index >= MAX_KEYS ))
	index = lp->TransmitKeyID - 1;
    
    erq->flags |= index + 1;
    
    /* Copy the key to the user buffer */
    erq->length = lp->DefaultKeys.key[index].len;
    memcpy(key, lp->DefaultKeys.key[index].key, erq->length);

    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_siwnickn(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *nickname)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    /* Validate the new value */
    if(data->length > HCF_MAX_NAME_LEN) {
        ret = -EINVAL;
	goto out;
    }
    
    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    memset(lp->StationName, 0, sizeof(lp->StationName));
    memcpy(lp->StationName, nickname, data->length);

    /* Commit the adapter parameters */
    wl_apply(lp);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwnickn(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *nickname)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int         ret = 0;
    int         status = -1;
    wvName_t    *pName;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Get the current station name */
    lp->ltvRecord.len = 1 + (sizeof(*pName)/sizeof(hcf_16));
    lp->ltvRecord.typ = CFG_CNF_OWN_NAME;
    status = hcf_get_info(&(lp->hcfCtx ), (LTVP)&(lp->ltvRecord));
    
    if(status == HCF_SUCCESS) {
        pName = (wvName_t *)&(lp->ltvRecord.u.u32);

        /* Endian translate the length */
        pName->length = __le16_to_cpu(pName->length);
        
	if (pName->length > IW_ESSID_MAX_SIZE)
	    ret = -EFAULT;
	else {
	    /* Copy the information into the user buffer */
	    data->length = pName->length;
	    memcpy(nickname, pName->name, pName->length);
	}
    } else
        ret = -EFAULT;

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_siwmode(struct net_device *dev,
		      struct iw_request_info *info,
		      __u32 *mode, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);
    
    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    switch(*mode) {
    case IW_MODE_ADHOC:
        /* When user requests ad-hoc, set IBSS mode! */
        lp->PortType         = 1;
        lp->CreateIBSS       = 1;
        lp->DownloadFirmware = WVLAN_DRV_MODE_STA;
        break;

    case IW_MODE_AUTO:
    case IW_MODE_INFRA:
        /* Both automatic and infrastructure set port to BSS/STA mode */
        lp->PortType         = 1;
        lp->CreateIBSS       = 0;
        lp->DownloadFirmware = WVLAN_DRV_MODE_STA;
        break;

#if (HCF_TYPE) & HCF_TYPE_AP   
    case IW_MODE_MASTER:
        /* Set BSS/AP mode */
        lp->PortType         = 1;
        lp->CreateIBSS       = 0;
        lp->DownloadFirmware = WVLAN_DRV_MODE_AP;
        break;
#endif /* (HCF_TYPE) & HCF_TYPE_AP */

    default:
        ret = -EINVAL;
	goto out_unlock;
    }
    
    /* Commit the adapter parameters */
    wl_go(lp);

    /* Send an event that mode has been set */ 
    wl_wext_event_mode(lp->dev);

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwmode(struct net_device *dev,
		      struct iw_request_info *info,
		      __u32 *mode, char *extra)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int status = -1;
    int     ret = 0;
    hcf_16  *pPortType;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Get the current port type */
    lp->ltvRecord.len = 1 + (sizeof(*pPortType)/sizeof(hcf_16));
    lp->ltvRecord.typ = CFG_CNF_PORT_TYPE;
    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
    
    if( status == HCF_SUCCESS ) {
	pPortType = (hcf_16 *)&(lp->ltvRecord.u.u32);
	*pPortType = __le16_to_cpu(*pPortType);

        switch(*pPortType) {
        case 1:
	case 13: /* ??? apparently this is set to that for Master */
            if(lp->DownloadFirmware == WVLAN_DRV_MODE_AP)
                *mode = IW_MODE_MASTER;
            else if(lp->CreateIBSS )
		*mode = IW_MODE_ADHOC;
	    else
		*mode = IW_MODE_INFRA;
            break;

	case 3:
            *mode = IW_MODE_ADHOC;
            break;

	default:
	    ret = -EFAULT;
	    break;
        }
    } else
        ret = -EFAULT;

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_siwpower(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_param *wrq, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (!wrq->disabled && (wrq->flags || wrq->value)) {
	ret = -EINVAL;
	goto out;
    }

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Set the power management state based on the 'disabled' value */
    lp->PMEnabled = !wrq->disabled;

    /* Commit the adapter parameters */
    wl_apply(lp);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwpower(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_param *rrq, char *extra)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    rrq->flags = 0;
    rrq->value = 0;

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    if(lp->PMEnabled)
        rrq->disabled = 0;
    else
        rrq->disabled = 1;

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwtxpow(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_param *rrq, char *extra)
{
    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

#ifdef USE_POWER_DBM
    rrq->value = RADIO_TX_POWER_DBM;
    rrq->flags = IW_TXPOW_DBM;
#else
    rrq->value = RADIO_TX_POWER_MWATT;
    rrq->flags = IW_TXPOW_MWATT;
#endif
    rrq->fixed = 1;
    rrq->disabled = 0;

    DBG_LEAVE( DbgInfo );
    return 0;
}

static int wl_siwrts (struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *rts, char *extra)
{
    int ret = 0;
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int rthr = rts->value;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );
    
    if(rts->fixed == 0) {
        ret = -EINVAL;
	goto out;
    }

#if WIRELESS_EXT > 8
    if(rts->disabled)
        rthr = 2347;
#endif /* WIRELESS_EXT > 8 */
    
    if((rthr < 256) || (rthr > 2347)) {
        ret = -EINVAL;
	goto out;
    }
    
    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    lp->RTSThreshold = rthr;
    wl_apply(lp);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_giwrts (struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *rts, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (wl_lock(lp, &flags))
	return -EBUSY;

    rts->value = lp->RTSThreshold;

    wl_unlock(lp, &flags);

#if WIRELESS_EXT > 8

    rts->disabled = (rts->value == 2347);

#endif /* WIRELESS_EXT > 8 */
    
    rts->fixed = 1;

    DBG_LEAVE( DbgInfo );
    return 0;
}

static int wl_siwrate(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *rrq, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int rate = 0;
    int ret = 0;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );


    /* Convert the bitrate to a required value. Rates are stored as:
            1 - Low Speed                   (1  Mb/sec)
            2 - Standard Speed              (2  Mb/sec)
            3 - Auto Rate Select (High)     (11 Mb/sec)
            4 - Medium Speed                (5  Mb/sec)
            5 - High Speed                  (11 Mb/sec)
            6 - Auto Rate Select (Standard) (2  Mb/sec)
            7 - Auto Rate Select (Medium)   (5  Mb/sec)
    */


    if(rrq->value > 0 && rrq->value <= 1*MEGABIT)
        rate = 1;
    else if(rrq->value > 1*MEGABIT && rrq->value <= 2*MEGABIT) {
        if(rrq->fixed)
            rate = 2;
        else
            rate = 6;
    } else if(rrq->value > 2*MEGABIT && rrq->value <= 5*MEGABIT) {
        if(rrq->fixed)
            rate = 4;
        else
            rate = 7;
    } else if(rrq->value > 5*MEGABIT && rrq->value <= 11*MEGABIT) {
        if(rrq->fixed)
            rate = 5;
        else
            rate = 3;
    } else if(rrq->fixed == 0)
        /* In this case, the user has not specified a bitrate, only the "auto"
           moniker. So, set the rate to 11Mb auto */
        rate = 3;
    else {
        ret = -EINVAL; 
	goto out;
    }

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Commit the adapter parameters */
    lp->TxRateControl = rate;
    wl_apply( lp );

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_giwrate(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *rrq, char *extra)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int     ret = 0;
    int     status = -1;
    __u16   *pTxRate;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Get the current transmit rate from the adapter */
    lp->ltvRecord.len = 1 + ( sizeof(*pTxRate)/sizeof(hcf_16));
    lp->ltvRecord.typ = CFG_CURRENT_TX_RATE;
    status = hcf_get_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));

    if (status == HCF_SUCCESS) {
        pTxRate = (__u16 *)&(lp->ltvRecord.u.u32);
        *pTxRate = __le16_to_cpu(*pTxRate);
        rrq->value = *pTxRate * MEGABIT;
    } else
        ret = -EFAULT; 

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

#if WIRELESS_EXT > 13
static int wl_siwscan(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_point *data, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int                 ret = 0;
    int                 status = -1;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* Set the completion state to FALSE */
    lp->probe_results.scan_complete = FALSE;

    /* Channels to scan */
    lp->ltvRecord.len       = 2;
    lp->ltvRecord.typ       = CFG_SCAN_CHANNELS_2GHZ; // need a define in MDD.h
    lp->ltvRecord.u.u16[0]  = __cpu_to_le16(0x7FFF);
    status = hcf_put_info(&(lp->hcfCtx), (LTVP)&( lp->ltvRecord));
    if (ret != HCF_SUCCESS) {
	ret = -EFAULT;
	goto out_unlock;
    }

    DBG_TRACE(DbgInfo, "CFG_SCAN_CHANNELS_2GHZ result: 0x%x\n", status);

    /* Set the SCAN_SSID to "ANY". Using this RID for scan prevents the need to
       disassociate from the network we are currently on */
    lp->ltvRecord.len       = 2;
    lp->ltvRecord.typ       = CFG_SCAN_SSID;
    lp->ltvRecord.u.u16[0]  = __cpu_to_le16(0);
    status = hcf_put_info(&(lp->hcfCtx), (LTVP)&(lp->ltvRecord));
    if (ret != HCF_SUCCESS) {
	ret = -EFAULT;
	goto out_unlock;
    }

    DBG_TRACE(DbgInfo, "CFG_SCAN_SSID to 'any' status: 0x%x\n", status);

    /* Initiate the scan */
    /* NOTE: Using HCF_ACT_SCAN has been removed, as using HCF_ACT_ACS_SCAN to
       retrieve probe responses must always be used to support WPA */
    status = hcf_action(&(lp->hcfCtx), HCF_ACT_ACS_SCAN);

    if( status == HCF_SUCCESS )
        DBG_TRACE(DbgInfo, "SUCCESSFULLY INITIATED SCAN...\n");
    else
	ret = -EFAULT;

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}

static int wl_giwscan(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_point *data, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int                 ret = 0;
    int                 count;
    char		*buf = extra;
    char                *end = extra + IW_SCAN_MAX_DATA;
    struct iw_event     iwe;
    PROBE_RESP          *probe_resp;
    hcf_8               msg[512];
    hcf_8               *wpa_ie;
    hcf_16              wpa_ie_len;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    /* If the scan is not done, tell the calling process to try again later */
    if( !lp->probe_results.scan_complete ) {
        ret = -EAGAIN;
	goto out_unlock;
    }

    DBG_TRACE(DbgInfo, "SCAN COMPLETE, Num of APs: %d\n", 
              lp->probe_results.num_aps);

    for (count = 0; count < lp->probe_results.num_aps; count++) {
        /* Reference the probe response from the table */
        probe_resp = (PROBE_RESP *)&lp->probe_results.ProbeTable[count];

        /* First entry MUST be the MAC address */
        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd                 = SIOCGIWAP;
        iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
        memcpy(iwe.u.ap_addr.sa_data, probe_resp->BSSID, ETH_ALEN);
        iwe.len                 = IW_EV_ADDR_LEN;
        buf = iwe_stream_add_event(buf, end, &iwe, IW_EV_ADDR_LEN);

        /* Use the mode to indicate if it's a station or AP */
        /* Won't always be an AP if in IBSS mode */
        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd = SIOCGIWMODE;
        if( probe_resp->capability & CAPABILITY_IBSS )
            iwe.u.mode = IW_MODE_INFRA;
        else
            iwe.u.mode = IW_MODE_MASTER;
        iwe.len = IW_EV_UINT_LEN;
        buf = iwe_stream_add_event(buf, end, &iwe, IW_EV_UINT_LEN);

        /* Any quality information */
        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd             = IWEVQUAL;
        iwe.u.qual.level    = dbm(probe_resp->signal);
        iwe.u.qual.noise    = dbm(probe_resp->silence);
        iwe.u.qual.qual     = iwe.u.qual.level - iwe.u.qual.noise;
        iwe.u.qual.updated  = lp->probe_results.scan_complete;
        iwe.len             = IW_EV_QUAL_LEN;
        buf = iwe_stream_add_event(buf, end, &iwe, IW_EV_QUAL_LEN);

        /* ESSID information */
        if (probe_resp->rawData[1] > 0) {
            memset(&iwe, 0, sizeof(iwe));
            iwe.cmd = SIOCGIWESSID;
            iwe.u.data.length = probe_resp->rawData[1];
            iwe.u.data.flags = 1;
            buf = iwe_stream_add_point(buf, end, &iwe, &probe_resp->rawData[2]);
        }

        /* Encryption Information */
        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd             = SIOCGIWENCODE;
        iwe.u.data.length   = 0;
        /* Check the capabilities field of the Probe Response to see if
           'privacy' is supported on the AP in question */
        if( probe_resp->capability & CAPABILITY_PRIVACY )
            iwe.u.data.flags |= IW_ENCODE_ENABLED;
        else
            iwe.u.data.flags |= IW_ENCODE_DISABLED;
        buf = iwe_stream_add_point(buf, end, &iwe, NULL);

        /* Frequency Info */
        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd = SIOCGIWFREQ;
        iwe.len = IW_EV_FREQ_LEN;
        iwe.u.freq.m = wl_parse_ds_ie(probe_resp);
        iwe.u.freq.e = 0;
        buf = iwe_stream_add_event(buf, end, &iwe, IW_EV_FREQ_LEN);


#if WIRELESS_EXT > 14 
        /* Custom info (Beacon Interval) */
        memset(&iwe, 0, sizeof(iwe));
        memset(msg, 0, sizeof(msg));
        iwe.cmd = IWEVCUSTOM;
        sprintf(msg, "beacon_interval=%d", probe_resp->beaconInterval);
        iwe.u.data.length = strlen( msg );
        buf = iwe_stream_add_point(buf, end, &iwe, msg);
        

        /* Custom info (WPA-IE) */
        wpa_ie = NULL;
        wpa_ie_len = 0;
        wpa_ie = wl_parse_wpa_ie(probe_resp, &wpa_ie_len);
        if( wpa_ie != NULL ) {
            memset(&iwe, 0, sizeof(iwe));
            memset(msg, 0, sizeof(msg));
            iwe.cmd = IWEVCUSTOM;
            sprintf(msg, "wpa_ie=%s", wl_print_wpa_ie( wpa_ie, wpa_ie_len));
            iwe.u.data.length = strlen(msg);
            buf = iwe_stream_add_point(buf, end, &iwe, msg);
        }

        /* Add other custom info in formatted string format as needed... */
#endif
    }

    data->length = buf - extra;

out_unlock:
    wl_unlock(lp, &flags);

out:
    DBG_LEAVE(DbgInfo);
    return ret;
}
#endif  // WIRELESS_EXT > 13


#if WIRELESS_EXT > 17

static int wl_siwauth(struct net_device *dev,
		      struct iw_request_info *info,
		      struct iw_param *data, char *extra)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int                 ret;
    int			iwa_idx = data->flags & IW_AUTH_INDEX;
    int			iwa_val = data->value;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    switch (iwa_idx) {
	case IW_AUTH_WPA_VERSION:
	    /* We do support WPA only; how should DISBLED be treated? */
	    if (iwa_val == IW_AUTH_WPA_VERSION_WPA)
		ret = 0;
	    else
		ret = -EINVAL;
	    break;

	case IW_AUTH_WPA_ENABLED:
	    DBG_TRACE( DbgInfo, "IW_AUTH_WPA_ENABLED: val = %d\n", iwa_val);
	    if (iwa_val)
		lp->EnableEncryption = 2;
	    else
		lp->EnableEncryption = 0;
	    ret = 0;
	    break;

	case IW_AUTH_TKIP_COUNTERMEASURES:
	    lp->driverEnable = !iwa_val;
	    if(lp->driverEnable)
		hcf_cntl(&(lp->hcfCtx), HCF_CNTL_ENABLE | HCF_PORT_0);
	    else
		hcf_cntl(&(lp->hcfCtx), HCF_CNTL_DISABLE | HCF_PORT_0);
	    ret = 0;
	    break;

	case IW_AUTH_DROP_UNENCRYPTED:
	    /* We do not actually do anything here, just to silence
	     * wpa_supplicant */
	    ret = 0;
	    break;

	default:
	    /* return an error */
	    ret = -EINVAL;
	    break;
    }

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int hermes_set_key(ltv_t *ltv, int alg, int key_idx, int set_tx,
			  u8 *seq, u8 *key, size_t key_len)
{
    int ret = -EINVAL;
    // int   count = 0;
    int   buf_idx = 0;
    hcf_8 tsc[] = { 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00 };

    /* Check the key index here; if 0, load as Pairwise Key, otherwise, load as
       a group key. Note that for the Hermes, the RIDs for group/pariwise keys
       are different from each other and different than the default WEP keys as
       well. */
    switch( alg )
    {
    case IW_ENCODE_ALG_TKIP:
#if 0
	/* Make sure that there is no data queued up in the firmware before
	   setting the TKIP keys. If this check is not performed, some data
	   may be sent out with incorrect MIC and cause synchronizarion
	   errors with the AP */
	/* Check every 1ms for 100ms */
	for( count = 0; count < 100; count++ )
	{
	    usleep( 1000 );

	    ltv.len = 2;
	    ltv.typ = 0xFD91;  // This RID not defined in HCF yet!!!
	    ltv.u.u16[0] = 0;

	    wl_get_info( sock, &ltv, ifname );
	    
	    if( ltv.u.u16[0] == 0 )
	    {
		break;
	    }
	}

	if( count == 100 )
	{
	    wpa_printf( MSG_DEBUG, "Timed out waiting for TxQ!" );
	}
#endif
            
	/* valid for Hermes-I only */
	ltv->len = 26;
	ltv->typ = CFG_ADD_TKIP_DEFAULT_KEY;

	/* Load the key Index */
	ltv->u.u16[buf_idx] = key_idx;

	/* If this is a Tx Key, set bit 8000 */
	if( set_tx )
	{
	    ltv->u.u16[buf_idx] |= 0x8000;
	}

	buf_idx += 2;

	/* Load the RSC */
	memcpy( &ltv->u.u8[buf_idx], seq, 8 );
	buf_idx += 8;

	/* Load the TKIP, TxMIC, and RxMIC keys in one shot, because in
	   CFG_ADD_TKIP_DEFAULT_KEY they are back-to-back */
	memcpy( &ltv->u.u8[buf_idx], key, key_len );
	buf_idx += key_len;

	/* Load the TSC */
	memcpy( &ltv->u.u8[buf_idx], tsc, 8 );

	ltv->u.u16[0] = CNV_INT_TO_LITTLE( ltv->u.u16[0] );
	ret = 0;

	break;

    case IW_ENCODE_ALG_WEP:
    case IW_ENCODE_ALG_CCMP:
	break;

    case IW_ENCODE_ALG_NONE:
	/* Valid for Hermes-I only */
	/* Clear the Group TKIP keys by index */
	ltv->len = 2;
	ltv->typ = CFG_REMOVE_TKIP_DEFAULT_KEY;

	ltv->u.u16[0] = key_idx;
	
	ltv->u.u16[0] = CNV_INT_TO_LITTLE( ltv->u.u16[0] );
	ret = 0;
	break;

    default:
	break;
    }

    return ret;
}

static int hermes_apply(struct wl_private *lp, ltv_t *ltv)
{
    int ret;
    
    hcf_action(&(lp->hcfCtx), HCF_ACT_INT_OFF);
    ret = hcf_put_info(&(lp->hcfCtx), (LTVP)ltv);
    hcf_action(&(lp->hcfCtx), HCF_ACT_INT_ON);
    return ret;
}

static int wl_siwencodeext (struct net_device *dev,
			    struct iw_request_info *info,
			    struct iw_point *erq, char *keybuf)
{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int                 ret;
    int	key_idx = (erq->flags&IW_ENCODE_INDEX) - 1;
    ltv_t ltv;
    struct iw_encode_ext *ext = (struct iw_encode_ext *)keybuf;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    if (sizeof(ext->rx_seq) != 8) {
	DBG_TRACE(DbgInfo, "rz_seq size mismatch\n");
	DBG_LEAVE(DbgInfo);
	return -EINVAL;
    }

    if (wl_lock(lp, &flags)) {
	ret = -EBUSY;
	goto out;
    }

    memset(&ltv, 0, sizeof(ltv));
    ret = hermes_set_key(&ltv, ext->alg, key_idx,
		ext->ext_flags&IW_ENCODE_EXT_SET_TX_KEY,
		ext->rx_seq, ext->key, ext->key_len);

    if (ret == 0)
	ret = hermes_apply(lp, &ltv);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return ret;
}

static int wl_siwgenie(struct net_device *dev,
		       struct iw_request_info *info,
		       struct iw_point *data, char *extra)

{
    struct wl_private *lp = wl_priv(dev);
    unsigned long flags;
    int   ret = 0;
    ltv_t ltv;

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER( DbgInfo );

    memset(&ltv, 0, sizeof(ltv));

    if (wl_lock(lp, &flags) == 0) {
	/* Hermes-I PSK hardcoded */
	lp->AuthKeyMgmtSuite = ltv.u.u16[0] = 2;
	ltv.u.u16[0]  = CNV_INT_TO_LITTLE( ltv.u.u16[0] );

	ret = hermes_apply(lp, &ltv);
	wl_unlock(lp, &flags);
    } else
	ret = -EBUSY

    DBG_LEAVE( DbgInfo );
    return ret;
}
#endif // WIRELESS_EXT > 17


static struct iw_statistics *wl_wireless_stats(struct net_device *dev)
{
    struct iw_statistics    *pStats;
    struct wl_private       *lp = wl_priv(dev);

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);
    DBG_PARAM(DbgInfo, "dev", "%s (0x%p)", dev->name, dev);

    pStats = NULL;

    /* Initialize the statistics */
    pStats                  = &( lp->wstats );
    pStats->qual.updated    = 0x00;

    if( !( lp->flags & WVLAN2_UIL_BUSY ))
    {
	CFG_COMMS_QUALITY_STRCT    *pQual;
	CFG_HERMES_TALLIES_STRCT tallies;
	int                         status;

	/* Update driver status */
	pStats->status = 0;

	/* Get the current link quality information */
	lp->ltvRecord.len = 1 + ( sizeof( *pQual ) / sizeof( hcf_16 ));
	lp->ltvRecord.typ = CFG_COMMS_QUALITY;

	status = hcf_get_info( &( lp->hcfCtx ), (LTVP)&( lp->ltvRecord ));

	if( status == HCF_SUCCESS )
	{
	    pQual = (CFG_COMMS_QUALITY_STRCT *)&( lp->ltvRecord );

#ifdef USE_DBM
	    pStats->qual.qual  = (u_char) CNV_LITTLE_TO_INT( pQual->coms_qual );
	    pStats->qual.level = (u_char) dbm( CNV_LITTLE_TO_INT( pQual->signal_lvl ));
	    pStats->qual.noise = (u_char) dbm( CNV_LITTLE_TO_INT( pQual->noise_lvl ));
#else
	    pStats->qual.qual = percent( CNV_LITTLE_TO_INT( pQual->coms_qual ),
					 HCF_MIN_COMM_QUALITY,
					 HCF_MAX_COMM_QUALITY );

	    pStats->qual.level = percent( CNV_LITTLE_TO_INT( pQual->signal_lvl ),
					  HCF_MIN_SIGNAL_LEVEL,
					  HCF_MAX_SIGNAL_LEVEL );

	    pStats->qual.noise = percent( CNV_LITTLE_TO_INT( pQual->noise_lvl ),
					  HCF_MIN_NOISE_LEVEL,
					  HCF_MAX_NOISE_LEVEL );
#endif /* USE_DBM */

	    pStats->qual.updated |= 0x07;
	}
	else
	{
	    memset( &( pStats->qual ), 0, sizeof( pStats->qual ));
	}

	/* Get the current tallies from the adapter */
	if(wl_get_tallies(lp, &tallies) == 0)
	{
	    /* No endian translation is needed here, as CFG_TALLIES is an
	       MSF RID; all processing is done on the host, not the card! */
	    /* bor: it is not clear what happens but apparently either
	     * values are composed of 2 half words or for whatever
	     * reasons are located in upper word. For now just truncate it,
	     * it provides enough info even wrapped */

	    pStats->discard.nwid = 0L;
	    pStats->discard.code = tallies.RxWEPUndecryptable >> 16;
	    pStats->discard.retries = tallies.TxRetryLimitExceeded >> 16;
	    pStats->discard.fragment = tallies.RxMsgInBadMsgFragments >> 16;
	    pStats->discard.misc = (tallies.TxDiscards >> 16) + 
				   (tallies.RxFCSErrors >> 16) + 
				   (tallies.RxDiscards_NoBuffer >> 16) +
				   (tallies.TxDiscardsWrongSA >> 16);
	}
	else
	{
	    memset( &( pStats->discard ), 0, sizeof( pStats->discard ));
	}
    }

    DBG_LEAVE( DbgInfo );
    return pStats;
}

static struct iw_statistics * wl_get_wireless_stats(struct net_device *dev)
{
    struct wl_private       *lp = wl_priv(dev);
    unsigned long           flags;
    struct iw_statistics    *pStats = NULL;  

    DBG_FUNC(__FUNCTION__);
    DBG_ENTER(DbgInfo);

    if (wl_lock(lp, &flags))
	goto out;

    hcf_action(&(lp->hcfCtx), HCF_ACT_INT_OFF);

#ifdef USE_RTS
    if( lp->useRTS == 1 )
    {
	DBG_TRACE( DbgInfo, "Skipping wireless stats, in RTS mode\n" );
	spin_unlock_irqrestore( &( lp->slock ), flags );
	
	DBG_LEAVE( DbgInfo );
	return NULL;
    }
#endif

    pStats = wl_wireless_stats(dev);

    hcf_action(&(lp->hcfCtx), HCF_ACT_INT_ON);

    wl_unlock(lp, &flags);

out:
    DBG_LEAVE( DbgInfo );
    return pStats;
}

void wl_spy_gather(struct net_device *dev, u_char *mac)
{
    int                     status;
    u_char                  stats[2];
    DESC_STRCT              desc[1];
    struct wl_private   *lp = wl_priv(dev);
    struct iw_quality wstats;

    /* shortcut */
    if (!lp->spy_data.spy_number)
	return;

    /* Gather wireless spy statistics: for each packet, compare the source 
       address with out list, and if match, get the stats. */
    memset(stats, 0, sizeof(stats));
    memset(desc, 0, sizeof(DESC_STRCT));

    desc[0].buf_addr        = stats;
    desc[0].BUF_SIZE        = sizeof(stats);
    desc[0].next_desc_addr  = 0;			// terminate list

    status = hcf_rcv_msg(&(lp->hcfCtx), desc, 0);

    if( status == HCF_SUCCESS ) {
	wstats.level = (u_char) dbm(stats[1]);
	wstats.noise = (u_char) dbm(stats[0]);
	wstats.qual  = wstats.level > wstats.noise ? wstats.level - wstats.noise : 0;
	wstats.updated = 7;
	wireless_spy_update(dev, mac, &wstats);
    }
}

/*******************************************************************************
 *	wl_wext_event_freq()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that the channel/freq
 *      configuration for a specific device has changed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_freq( struct net_device *dev )
{
#if WIRELESS_EXT > 13
    union iwreq_data wrqu;
    struct wl_private *lp = wl_priv(dev);
    /*------------------------------------------------------------------------*/


    memset( &wrqu, 0, sizeof( wrqu ));

    wrqu.freq.m = lp->Channel;
    wrqu.freq.e = 0;

    wireless_send_event( dev, SIOCSIWFREQ, &wrqu, NULL );
#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_mode()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that the mode of operation
 *      for a specific device has changed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_mode( struct net_device *dev )
{
    #if WIRELESS_EXT > 13
    union iwreq_data wrqu;
    struct wl_private *lp = wl_priv(dev);
    /*------------------------------------------------------------------------*/


    memset( &wrqu, 0, sizeof( wrqu ));

    if( lp->DownloadFirmware == WVLAN_DRV_MODE_STA )
        wrqu.mode = IW_MODE_INFRA;
    else
        wrqu.mode = IW_MODE_MASTER;

	wireless_send_event( dev, SIOCSIWMODE, &wrqu, NULL );
#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_essid()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that the ESSID configuration for 
 *      a specific device has changed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_essid( struct net_device *dev )
{
#if WIRELESS_EXT > 13
    union iwreq_data wrqu;
    struct wl_private *lp = wl_priv(dev);
    /*------------------------------------------------------------------------*/

    
    memset( &wrqu, 0, sizeof( wrqu ));

    /* Fill out the buffer. Note that the buffer doesn't actually contain the
       ESSID, but a pointer to the contents. In addition, the 'extra' field of
       the call to wireless_send_event() must also point to where the ESSID 
       lives */
    wrqu.essid.length  = strlen( lp->NetworkName );
    wrqu.essid.pointer = (caddr_t)lp->NetworkName;
    wrqu.essid.flags   = 1;

    wireless_send_event( dev, SIOCSIWESSID, &wrqu, lp->NetworkName );
#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_encode()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that the encryption configuration
 *      for a specific device has changed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_encode( struct net_device *dev )
{
#if WIRELESS_EXT > 13
    union iwreq_data wrqu;
    struct wl_private *lp = wl_priv(dev);
    int index = 0;
    /*------------------------------------------------------------------------*/

    
    memset( &wrqu, 0, sizeof( wrqu ));

    if( lp->EnableEncryption == 0 )
    {
        wrqu.encoding.flags = IW_ENCODE_DISABLED;
    }
    else
    {
        wrqu.encoding.flags |= lp->TransmitKeyID;

        index = lp->TransmitKeyID - 1;

        /* Only set IW_ENCODE_RESTRICTED/OPEN flag using lp->ExcludeUnencrypted
           if we're in AP mode */
#if (HCF_TYPE) & HCF_TYPE_AP

        if( lp->DownloadFirmware == WVLAN_DRV_MODE_AP )
        {
            if( lp->ExcludeUnencrypted )
            {
                wrqu.encoding.flags |= IW_ENCODE_RESTRICTED;
            }
            else
            {
                wrqu.encoding.flags |= IW_ENCODE_OPEN;
            }
        }

#endif  // HCF_TYPE_AP

        /* Only provide the key if permissions allow */
        if( capable( CAP_NET_ADMIN ))
        {
            wrqu.encoding.pointer = (caddr_t)lp->DefaultKeys.key[index].key;
            wrqu.encoding.length  = lp->DefaultKeys.key[index].len;
        }
        else
        {
            wrqu.encoding.flags |= IW_ENCODE_NOKEY;
        }
    }

    wireless_send_event( dev, SIOCSIWENCODE, &wrqu,
		     lp->DefaultKeys.key[index].key );
#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_ap()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that the device has been
 *      associated to a new AP.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_ap( struct net_device *dev )
{
#if WIRELESS_EXT > 13
	union iwreq_data wrqu;
    struct wl_private *lp = wl_priv(dev);
    int status;
    /*------------------------------------------------------------------------*/


    /* Retrieve the WPA-IEs used by the firmware and sned an event. We must send 
       this event BEFORE sending the association event, as there are timing
       issues with the hostap supplicant. The supplicant will attempt to process
       an EAPOL-Key frame from an AP before receiving this information, which
       is required properly process the said frame. */
    wl_wext_event_assoc_ie( dev );

    /* Get the BSSID */
    lp->ltvRecord.typ = CFG_CURRENT_BSSID;
    lp->ltvRecord.len = 4;

    status = hcf_get_info( &( lp->hcfCtx ), (LTVP)&( lp->ltvRecord ));
    if( status == HCF_SUCCESS )
    {
        memset( &wrqu, 0, sizeof( wrqu ));

        memcpy( wrqu.addr.sa_data, lp->ltvRecord.u.u8, ETH_ALEN );
	    wrqu.addr.sa_family = ARPHRD_ETHER;
	    wireless_send_event( dev, SIOCGIWAP, &wrqu, NULL );
    }

#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_scan_complete()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that a request for a network scan
 *      has completed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_scan_complete( struct net_device *dev )
{
#if WIRELESS_EXT > 13
    union iwreq_data wrqu;
    /*------------------------------------------------------------------------*/


	memset( &wrqu, 0, sizeof( wrqu ));

    wrqu.addr.sa_family = ARPHRD_ETHER;
	wireless_send_event( dev, SIOCGIWSCAN, &wrqu, NULL );
#endif /* WIRELESS_EXT > 13 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_new_sta()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that an AP has registered a new
 *      station.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_new_sta( struct net_device *dev )
{
#if WIRELESS_EXT > 14
	union iwreq_data wrqu;
    /*------------------------------------------------------------------------*/


	memset( &wrqu, 0, sizeof( wrqu ));

    /* Send the station's mac address here */
    memcpy( wrqu.addr.sa_data, dev->dev_addr, ETH_ALEN );
	wrqu.addr.sa_family = ARPHRD_ETHER;
	wireless_send_event( dev, IWEVREGISTERED, &wrqu, NULL );
#endif /* WIRELESS_EXT > 14 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_expired_sta()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that an AP has deregistered a
 *      station.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_expired_sta( struct net_device *dev )
{
#if WIRELESS_EXT > 14
	union iwreq_data wrqu;
    /*------------------------------------------------------------------------*/


	memset( &wrqu, 0, sizeof( wrqu ));

	memcpy( wrqu.addr.sa_data, dev->dev_addr, ETH_ALEN );
	wrqu.addr.sa_family = ARPHRD_ETHER;
	wireless_send_event( dev, IWEVEXPIRED, &wrqu, NULL );
#endif /* WIRELESS_EXT > 14 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_mic_failed()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event that MIC calculations failed.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_mic_failed( struct net_device *dev )
{
#if WIRELESS_EXT > 14
    char               msg[512];
	union iwreq_data   wrqu;
    struct wl_private *lp = wl_priv(dev);
    int                key_idx;
    char              *addr1;
    char              *addr2;
    WVLAN_RX_WMP_HDR  *hdr; 
    /*------------------------------------------------------------------------*/


    key_idx = lp->lookAheadBuf[HFS_STAT+1] >> 3;
    key_idx &= 0x03;

    /* Cast the lookahead buffer into a RFS format */
    hdr = (WVLAN_RX_WMP_HDR *)&lp->lookAheadBuf[HFS_STAT];

    /* Cast the addresses to byte buffers, as in the above RFS they are word
       length */
    addr1 = (char *)hdr->address1;
    addr2 = (char *)hdr->address2;

    DBG_PRINT( "MIC FAIL - KEY USED : %d, STATUS : 0x%04x\n", key_idx,
               hdr->status );

	memset( &wrqu, 0, sizeof( wrqu ));
    memset( msg, 0, sizeof( msg ));


    /* Becuase MIC failures are not part of the Wireless Extensions yet, they
       must be passed as a string using an IWEVCUSTOM event. In order for the 
       event to be effective, the string format must be known by both the
       driver and the supplicant. The following is the string format used by the
       hostap project's WPA supplicant, and will be used here until the Wireless
       Extensions interface adds this support:
       
       MLME-MICHAELMICFAILURE.indication(keyid=# broadcast/unicast addr=addr2)
   */

    /* NOTE: Format of MAC address (using colons to seperate bytes) may cause
             a problem in future versions of the supplicant, if they ever
             actually parse these parameters */
    sprintf( msg, "MLME-MICHAELMICFAILURE.indication(keyid=%d %scast addr="
             "%s)", key_idx, addr1[0] & 0x01 ? "broad" : "uni",
             DbgHwAddr( addr2 ));
    wrqu.data.length = strlen( msg );
	wireless_send_event( dev, IWEVCUSTOM, &wrqu, msg );
#endif /* WIRELESS_EXT > 14 */

    return;
}
/*============================================================================*/




/*******************************************************************************
 *	wl_wext_event_assoc_ie()
 *******************************************************************************
 *
 *  DESCRIPTION:
 *
 *      This function is used to send an event containing the WPA-IE generated
 *      by the firmware in an association request.
 *      
 *
 *  PARAMETERS:
 *
 *      dev - the network device for which this event is to be issued
 *
 *  RETURNS:
 *
 *      N/A
 *
 ******************************************************************************/
void wl_wext_event_assoc_ie( struct net_device *dev )
{
#if WIRELESS_EXT > 14
    char               msg[512];
	union iwreq_data   wrqu;
    struct wl_private *lp = wl_priv(dev);
    int status;
    PROBE_RESP         data;
    hcf_16             length;
    hcf_8              *wpa_ie;
    /*------------------------------------------------------------------------*/


    memset( &wrqu, 0, sizeof( wrqu ));
    memset( msg, 0, sizeof( msg ));

    /* Retrieve the Association Request IE */
    lp->ltvRecord.len = 45;
    lp->ltvRecord.typ = CFG_CUR_ASSOC_REQ_INFO;

    status = hcf_get_info( &( lp->hcfCtx ), (LTVP)&( lp->ltvRecord ));
    if( status == HCF_SUCCESS )
    {
        length = 0;
        memcpy( &data.rawData, &( lp->ltvRecord.u.u8[1] ), 88 );
        wpa_ie = wl_parse_wpa_ie( &data, &length );

        /* Becuase this event (Association WPA-IE) is not part of the Wireless
        Extensions yet, it must be passed as a string using an IWEVCUSTOM event.
        In order for the event to be effective, the string format must be known
        by both the driver and the supplicant. The following is the string format
        used by the hostap project's WPA supplicant, and will be used here until 
        the Wireless Extensions interface adds this support:
           
        ASSOCINFO(ReqIEs=WPA-IE RespIEs=WPA-IE)
        */

        if( length != 0 )
        {
            sprintf( msg, "ASSOCINFO(ReqIEs=%s)", wl_print_wpa_ie( wpa_ie, length ));
	        wrqu.data.length = strlen( msg );
            wireless_send_event( dev, IWEVCUSTOM, &wrqu, msg );
        }
    }
#endif /* WIRELESS_EXT > 14 */

    return;
}
/*============================================================================*/
/* Structures to export the Wireless Handlers */

static const iw_handler wl_handler[] =
{
	(iw_handler) wl_siwcommit,		/* SIOCSIWCOMMIT */
	(iw_handler) wl_giwname,		/* SIOCGIWNAME */
	(iw_handler) NULL,			/* SIOCSIWNWID */
	(iw_handler) NULL,			/* SIOCGIWNWID */
	(iw_handler) wl_siwfreq,		/* SIOCSIWFREQ */
	(iw_handler) wl_giwfreq,		/* SIOCGIWFREQ */
	(iw_handler) wl_siwmode,		/* SIOCSIWMODE */
	(iw_handler) wl_giwmode,		/* SIOCGIWMODE */
	(iw_handler) wl_siwsens,		/* SIOCSIWSENS */
	(iw_handler) wl_giwsens,		/* SIOCGIWSENS */
	(iw_handler) NULL ,		/* SIOCSIWRANGE */
	(iw_handler) wl_giwrange,		/* SIOCGIWRANGE */
	(iw_handler) NULL ,		/* SIOCSIWPRIV */
	(iw_handler) NULL /* kernel code */,	/* SIOCGIWPRIV */
	(iw_handler) NULL ,		/* SIOCSIWSTATS */
	(iw_handler) NULL /* kernel code */,	/* SIOCGIWSTATS */
	iw_handler_set_spy,			/* SIOCSIWSPY */
	iw_handler_get_spy,			/* SIOCGIWSPY */
	NULL,				/* SIOCSIWTHRSPY */
	NULL,				/* SIOCGIWTHRSPY */
	(iw_handler) NULL,		/* SIOCSIWAP */
#if (HCF_TYPE) & HCF_TYPE_STA
	(iw_handler) wl_giwap,			/* SIOCGIWAP */
#else
	(iw_handler) NULL,		/* SIOCGIWAP */
#endif
	(iw_handler) NULL,		/* SIOCSIWMLME */
	(iw_handler) wl_giwaplist,		/* SIOCGIWAPLIST */
	(iw_handler) wl_siwscan,		/* SIOCSIWSCAN */
	(iw_handler) wl_giwscan,		/* SIOCGIWSCAN */
	(iw_handler) wl_siwessid,		/* SIOCSIWESSID */
	(iw_handler) wl_giwessid,		/* SIOCGIWESSID */
	(iw_handler) wl_siwnickn,		/* SIOCSIWNICKN */
	(iw_handler) wl_giwnickn,		/* SIOCGIWNICKN */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) wl_siwrate,		/* SIOCSIWRATE */
	(iw_handler) wl_giwrate,		/* SIOCGIWRATE */
	(iw_handler) wl_siwrts,			/* SIOCSIWRTS */
	(iw_handler) wl_giwrts,			/* SIOCGIWRTS */
	(iw_handler) NULL,		/* SIOCSIWFRAG */
	(iw_handler) NULL,		/* SIOCGIWFRAG */
	(iw_handler) NULL,		/* SIOCSIWTXPOW */
	(iw_handler) wl_giwtxpow,		/* SIOCGIWTXPOW */
	(iw_handler) NULL,		/* SIOCSIWRETRY */
	(iw_handler) NULL,		/* SIOCGIWRETRY */
	(iw_handler) wl_siwencode,		/* SIOCSIWENCODE */
	(iw_handler) wl_giwencode,		/* SIOCGIWENCODE */
	(iw_handler) wl_siwpower,		/* SIOCSIWPOWER */
	(iw_handler) wl_giwpower,		/* SIOCGIWPOWER */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) wl_siwgenie,		/* SIOCSIWGENIE */
	(iw_handler) NULL,		/* SIOCGIWGENIE */
	(iw_handler) wl_siwauth,		/* SIOCSIWAUTH */
	(iw_handler) NULL,		/* SIOCGIWAUTH */
	(iw_handler) wl_siwencodeext,		/* SIOCSIWENCODEEXT */
	(iw_handler) NULL,		/* SIOCGIWENCODEEXT */
	(iw_handler) NULL,		/* SIOCSIWPMKSA */
	(iw_handler) NULL,			/* -- hole -- */
};

static const iw_handler wl_private_handler[] =
{						/* SIOCIWFIRSTPRIV + */
	(iw_handler) NULL,			/* 0: SIOCSIWNETNAME */
	wvlan_get_netname,			/* 1: SIOCGIWNETNAME */
	(iw_handler) NULL,			/* 2: SIOCSIWSTANAME */
	(iw_handler) NULL,			/* 3: SIOCGIWSTANAME */
	(iw_handler) NULL,			/* 4: SIOCSIWPORTTYPE */
	wvlan_get_porttype,			/* 5: SIOCGIWPORTTYPE */
};

struct iw_priv_args wl_priv_args[] = {
    {SIOCSIWNETNAME,
     IW_PRIV_TYPE_CHAR | HCF_MAX_NAME_LEN, 0, "snetwork_name" },
    {SIOCGIWNETNAME,
     0, IW_PRIV_TYPE_CHAR | HCF_MAX_NAME_LEN, "gnetwork_name" },
    {SIOCSIWSTANAME,
     IW_PRIV_TYPE_CHAR | HCF_MAX_NAME_LEN, 0, "sstation_name" },
    {SIOCGIWSTANAME,
     0, IW_PRIV_TYPE_CHAR | HCF_MAX_NAME_LEN, "gstation_name" },
    {SIOCSIWPORTTYPE,
     IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "sport_type" },
    {SIOCGIWPORTTYPE,
     0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "gport_type" },
};

const struct iw_handler_def wl_iw_handler_def =
{
	.num_private	= sizeof(wl_private_handler) / sizeof(iw_handler),
	.private	= (iw_handler *) wl_private_handler,
	.private_args	= (struct iw_priv_args *) wl_priv_args,
	.num_private_args = sizeof(wl_priv_args) / sizeof(struct iw_priv_args),
	.num_standard	= sizeof(wl_handler) / sizeof(iw_handler),
	.standard	= (iw_handler *) wl_handler,
	.get_wireless_stats = wl_get_wireless_stats,
};

#endif // WIRELESS_EXT
